module cu(M, CLK, res, PCo, SPo, ACCo, PCl, SPl, ACCl, OUTo, ALUc
           , RD, WR, MARl, CFl, ZFl, state,IR);
  input [7:0] M;
  input CLK, res;

  output PCo, SPo, ACCo, PCl, SPl, ACCl, OUTo;
  output [4:0] ALUc;
  output RD, WR, MARl;
  output CFl, ZFl;
  output [7:0] state;
  output [7:0] IR;

  reg start;
  reg [7:0] IR;
  reg PCo, SPo, ACCo, PCl, SPl, ACCl, OUTo;
  reg [4:0] ALUc;
  reg RD, WR, MARl;
  reg IRl;
  reg CFl, ZFl;

  reg [7:0] state;
  reg [7:0] next_state;

  reg HLT;
  reg ERR;

`define RESET 		8'h00
`define FETCH_0		8'h01
`define FETCH_1		8'h02
`define FETCH_2		8'h03
`define DECODE_0	8'h04
`define ACC_0		8'h05
`define LSP_SSP_0	8'h06
`define RET_POP_0	8'h07
`define RET_POP_1	8'h08
`define RET_POP_2	8'h09
`define PUSH_0		8'h0A
`define PUSH_1		8'h0B
`define TBYTE_PRE_0	8'h0C
`define TBYTE_PRE_1	8'h0D
`define TACC_0		8'h0E
`define	TACC_1		8'h0F
`define	TACC_2		8'h10
`define	TACC_3		8'h11
`define	JC_JZ_0		8'h12
`define	JMP_CALL_0	8'h13
`define	JMP_CALL_1	8'h14
`define	JMP_CALL_2	8'h15
`define	JMP_CALL_3	8'h16
`define	JMP_CALL_4	8'h17
`define	JMP_CALL_5	8'h18
`define	JMP_CALL_6	8'h19
`define	ST_0		8'h1A
`define	ST_1		8'h1B
`define	ST_2		8'h1C
`define	ST_3		8'h1D
`define OUT_0		8'h1E


  always @(posedge CLK)
    begin
      if (!res)
        begin
           state = `RESET;
        end else
        begin
            if(!HLT) state = next_state;
            if(IRl)  IR=M;
        end
    end


    always @( negedge CLK )
      case (state)
// RESET
        `RESET:
            begin
               PCo = 0;
               SPo = 0;
               ACCo = 0;
               PCl = 0;
               SPl = 0;
               ACCl = 0;
               ALUc = 0;
               RD = 0;
               WR = 0;
               MARl = 0;
               IRl = 0;
               CFl = 0;
               ZFl = 0;
               HLT = 0;
               next_state = `FETCH_0;
               ERR = 0;
               OUTo = 0;
            end
// FETCH
        `FETCH_0:
            begin
               PCo=1; ALUc=5'b00000; MARl=1;
               next_state=`FETCH_1;
            end
        `FETCH_1:
            begin
             /*PCo=0; ALUc=5'b00000;*/ MARl=0;
             /*PCo=1;*/ALUc=5'b10000; PCl=1;
               next_state=`FETCH_2;
            end
        `FETCH_2:
            begin
               PCo=0; ALUc=5'b00000; PCl=0;
               RD=1;IRl=1;
               next_state=`DECODE_0;
            end
// DECODE
        `DECODE_0:
            begin
              RD=0;IRl=0;
              if(IR[7] == 1'b0)
            // 1 byte operation
                begin
               // ACC operation
                    if(IR[7:6] == 2'b01)
	                   begin
	                      ACCo=1;ALUc={1'b1,IR[3:0]};ACCl=~&IR[3:1];CFl=IR[4];ZFl=IR[5];
                          next_state=`ACC_0;
                       end
               // LSP,SSP
                    else if(IR[7:2] == 6'b000001)
	                   begin
	                      ACCo=IR[0];SPo=IR[1];/*ALUc=5'b00000;*/ACCl=IR[1];SPl=IR[0];
                          next_state=`LSP_SSP_0;
                       end
               // RET, POP
	                else if(IR[7:2] == 6'b000010)
	                   begin
                          SPo=1;/*ALUc=5'b00000;*/MARl=1;
                          next_state=`RET_POP_0;
                       end
                    else case(IR)
                     8'h0c:begin
                    // PUSH
                         SPo=1;ALUc=5'b10001;SPl=1;MARl=1;
                         next_state=`PUSH_0;
                      end
                      8'h0e:begin
                    // OUT
                          ACCo=1;OUTo=1;
                          next_state=`OUT_0;
                      end
                      8'h0f:begin
                    // HLT
                          HLT=1;
                          next_state=`RESET;
                      end
                  endcase
	          end
            // 2 byte operation
              else if(IR[7] == 1'b1)
                 begin
                    PCo=1;/*ALUc=5'b00000;*/MARl=1;
                    next_state=`TBYTE_PRE_0;
                 end
	        end
// ACC operation
        `ACC_0:
            begin
               ACCo=0;ALUc=5'b00000;ACCl=0;CFl=0;ZFl=0;
               next_state=`FETCH_0;
            end
// LSP,SSP
        `LSP_SSP_0:
           begin
               ACCo=0;SPo=0;ALUc=5'b00000;ACCl=0;SPl=0;
               next_state=`FETCH_0;
           end
// RET, POP
        `RET_POP_0:
           begin
               SPo=0;/*ALUc=5'b00000;*/MARl=0;
               RD=1;ALUc=5'b00001;PCl=IR[0];ACCl=IR[1];
               next_state=`RET_POP_1;
           end
        `RET_POP_1:
           begin
               RD=0;/*ALUc=5'b00000;*/PCl=0;ACCl=0;
               SPo=1;ALUc=5'b10000;SPl=1;
               next_state=`RET_POP_2;
           end
        `RET_POP_2:
           begin
               SPo=0;ALUc=5'b00000;SPl=0;
               next_state=`FETCH_0;
           end
// PUSH
        `PUSH_0:
           begin
               SPo=0;/*ALUc=5'b00000;SPl=0;*/MARl=0;
               ACCo=1;ALUc=5'b00000;WR=1;
               next_state=`PUSH_1;
           end
        `PUSH_1:
           begin
               ACCo=0;ALUc=5'b00000;WR=0;
               next_state=`FETCH_0;
           end
// 2 byte operation
        `TBYTE_PRE_0:
           begin
               /*PCo=0; ALUc=5'b00000;*/MARl=0;
               PCo=1; ALUc=5'b10000; PCl=1;
               next_state=`TBYTE_PRE_1;
           end
        `TBYTE_PRE_1: // ok
           begin
           // ACC operation
               if(IR[7:6] == 2'b10)
                   begin
               //repeat(IR[5:4])
                      PCo=0;/*ALUc=5'b00000;*/ PCl=0;
                      case(IR[5:4])
                       0:begin
                            RD=1;ACCo=1;ALUc={1'b0,IR[3:0]};
                            ACCl=~&IR[2:0];CFl=IR[3];ZFl=|IR[3:2];
                            next_state=`TACC_3;
                         end
                       1:begin
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`TACC_2;
                         end
                       2:begin
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`TACC_1;
                         end
                       3:begin
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`TACC_0;
                         end
                      endcase
                   end
           // JC, JZ
	           else if(IR[7:3] == 5'b11000)
	               begin
                      PCo=0;/*ALUc=5'b00000; PCl=0;*/
                      RD=1;PCo=1;ALUc={1'b0,IR[3:0]};PCl=1;
                      next_state=`JC_JZ_0;
                   end
           // JMP, CALL
               else if(IR[7:3] == 5'b11001)
                   begin
                      if(IR[2] == 1'b1)
                         begin
                            PCo=0;/*ALUc=5'b00000;*/PCl=0;
                            SPo=1;ALUc=5'b10001;SPl=1;MARl=1;
                            next_state=`JMP_CALL_0;
                         end
                      case(IR[1:0])
                       0: begin
                             PCo=0;/*ALUc=5'b00000; PCl=0;*/
                             RD=1;ALUc=5'b00001;PCl=1;
                             next_state=`JMP_CALL_6;
                          end
                       1: begin
                             PCo=0;/*ALUc=5'b00000;*/PCl=0;
                             RD=1;ALUc=5'b00001;MARl=1;
                             next_state=`JMP_CALL_5;
                          end 
                       2: begin
                             PCo=0;/*ALUc=5'b00000;*/PCl=0;
                             RD=1;ALUc=5'b00001;MARl=1;
                             next_state=`JMP_CALL_4;
                          end 
                       3: begin
                             PCo=0;/*ALUc=5'b00000;*/PCl=0;
                             RD=1;ALUc=5'b00001;MARl=1;
                             next_state=`JMP_CALL_3;
                          end
                      endcase 
                   end
           // ST
	           else if(IR[7:2] == 6'b110100)
                   begin
                      case(IR[1:0])
                       0:begin
                            PCo=0;/*ALUc=5'b00000;*/PCl=0;
                            ACCo=1;ALUc=5'b00000;WR=1;
                            next_state=`ST_3;
                         end
                       1:begin
                            PCo=0;/*ALUc=5'b00000;*/PCl=0;
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`ST_2;
                         end
                       2:begin
                            PCo=0;/*ALUc=5'b00000;*/PCl=0;
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`ST_1;
                         end
                       3:begin
                            PCo=0;/*ALUc=5'b00000;*/PCl=0;
                            RD=1;ALUc=5'b00001;MARl=1;
                            next_state=`ST_0;
                         end
                      endcase
                   end
           // Illengal value
               else
                   begin
                      PCo=0; ALUc=5'b00000; PCl=0;
                      ERR=1; HLT=1;
                      next_state=`FETCH_0;
                   end
           // else ...more operation
           end
// 2Byte ACC
        `TACC_0:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`TACC_1;
           end
        `TACC_1:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`TACC_2;
           end
        `TACC_2:
           begin
            /*RD=0;ALUc=5'b00000;*/MARl=0;
              RD=1;ACCo=1;ALUc={1'b0,IR[3:0]};ACCl=~&IR[2:0];CFl=IR[3];ZFl=|IR[3:2];
              next_state=`TACC_3;
           end
        `TACC_3:
           begin
              RD=0;ACCo=0;ALUc=5'b00000;ACCl=0;CFl=0;ZFl=0;
              next_state=`FETCH_0;
           end
// JC,JZ
        `JC_JZ_0:
           begin
              RD=0;PCo=0;ALUc=5'b00000;PCl=0;
              next_state=`FETCH_0;
           end
// JMP,CALL
        `JMP_CALL_0:
           begin
              SPo=0;/*ALUc=5'b00000;*/SPl=0;MARl=0;
              PCo=1;ALUc=5'b00000;WR=1;
              next_state=`JMP_CALL_1;
           end
        `JMP_CALL_1:
           begin
              /*PCo=0;ALUc=5'b00000;*/WR=0;
              PCo=1;ALUc=5'b10001;MARl=1;
              next_state=`JMP_CALL_2;
           end
        `JMP_CALL_2:
           begin
              PCo=0;ALUc=5'b00000;MARl=0;
              next_state=`JMP_CALL_3;
           end
        `JMP_CALL_3:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`JMP_CALL_4;
           end
        `JMP_CALL_4:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`JMP_CALL_5;
           end
        `JMP_CALL_5:
           begin
              /*RD=0;ALUc=5'b00000;*/MARl=0;
              RD=1;ALUc=5'b00001;PCl=1;
              next_state=`JMP_CALL_6;
           end
        `JMP_CALL_6:
           begin
              RD=0;ALUc=5'b00000;PCl=0;
              next_state=`FETCH_0;
           end
// ST
        `ST_0:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`ST_1;
           end
        `ST_1:
           begin
              /*RD=0;ALUc=5'b00000;MARl=0;*/
              RD=1;ALUc=5'b00001;MARl=1;
              next_state=`ST_2;
           end
        `ST_2:
           begin
              RD=0;/*ALUc=5'b00000;*/MARl=0;
              ACCo=1;ALUc=5'b00000;WR=1;
              next_state=`ST_3;
           end
        `ST_3:
           begin
              ACCo=0;ALUc=5'b00000;WR=0;
              next_state=`FETCH_0;
           end
        `OUT_0:
           begin
              OUTo=0;ACCo=0;
              next_state=`FETCH_0;
           end
     endcase
endmodule
